(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklmlc.ml,v 1.14.2.1 2011-09-21 13:08:03 fclement Exp $ *)

(* This is the sklml compiler *)

open Install_data;;

type semantics_mode = Seq | Par;;

type compilation_options = {
  mode : semantics_mode;
  pp_cmd : string;
  ocamlc_options : string list;
};;

type compilation_mode =
  | Byt
  | Bin
  | Undefined
;;

let dir_of_semantics_mode = function
  | Seq -> mk_lib_subdir "seq"
  | Par -> mk_lib_subdir "par"
;;

let guess_with_exts e1 e2 opts =
  let flip_check_suffix suff fn = Filename.check_suffix fn suff in
  if List.exists (flip_check_suffix e1) opts
  then Byt
  else if List.exists (flip_check_suffix e2) opts
    then Bin
    else Undefined
;;

let guess_with_libs = guess_with_exts ".cma" ".cmxa";;
let guess_with_objs = guess_with_exts ".cmo" ".cmx";;

let guess_compiler_and_libs opts =
  let compilation_mode = List.fold_left
    (fun m f -> if m = Undefined then f opts else m)
    Undefined
    [ guess_with_objs; guess_with_libs; ] in
  let (compiler, lib_suffix) =
    match compilation_mode with
    | Bin -> (ocaml_compiler.bin, ".cmxa")
    | Byt | Undefined -> (ocaml_compiler.byt, ".cma") in
  let libs =
    if not (List.exists (( = ) "-c") opts)
    then List.map (fun x -> x ^ lib_suffix)
      [ "bigarray"; "unix"; "ulib"; "parlib"; "sklml"; "sklml_extra" ]
    else [] in
  (compiler, libs)
;;

let unwords = String.concat " ";;

let mk_command { mode = m; pp_cmd = pp_cmd; ocamlc_options = caml_options; } =
  let include_dirs = [ mk_lib_subdir ""; dir_of_semantics_mode m; "+ulib" ] in
  let include_opts = List.flatten
    (List.map (fun d -> [ "-I"; d; ]) include_dirs) in
  let pp_opt = [ "-pp"; Printf.sprintf "'%s'" pp_cmd; ] in
  let generic_opts = include_opts @ pp_opt in
  let (compiler, sklml_libs) = guess_compiler_and_libs caml_options in
  unwords ([compiler] @ generic_opts @ sklml_libs @ caml_options )
;;

let fail msg =
  Printf.eprintf "sklmlc: %s\n" msg; exit 1
;;

let parse_argv args =
  let (get_annons, accum_annons) =
    let annons = ref [] in
    (fun () -> !annons),
    (fun px s -> annons := !annons @ [px ^ s])
  and (get_pp, set_pp) =
    let pp = ref camlp45_pa_skl in
    (fun () -> !pp),
    (fun s -> pp := s)
  and (get_mode, set_mode) =
    let m = ref None in
    (fun () -> !m),
    (fun s -> match s with
      | "seq" -> m := Some Seq
      | "par" -> m := Some Par
      | _ -> assert false
    )
  and (get_debug, set_debug) =
    let d = ref false in
    (fun () -> !d),
    (fun () -> d := true)
  in
  let speclist = Arg.align
    [ "-mode", Arg.Symbol (["seq"; "par"], set_mode),
      "MODE Sets the compilation mode, must be 'par' or 'seq'"
    ; "-debug", Arg.Unit set_debug,
      " Toggle display of debug informations during compilation"
    ; "-ext", Arg.String (accum_annons "-"),
      " Internal use only"
    ; "-pp", Arg.String set_pp,
      "PP Sets the location of the camlp4/5 parser module to use"
    ] in
  let keys =
    [ "-help"; "--help"; ] @ List.map (fun (x, _, _) -> x) speclist in
  let args = Array.of_list
    (Array.fold_left (fun l s ->
      if not (List.mem s keys) then
        let key = String.sub s 1 (String.length s - 1) in
        if s.[0] = '-'
        then l @ [ "-ext"; key ]
        else l @ [s]
      else l @ [s]) [] args) in
  try
    Arg.parse_argv args speclist (accum_annons "")
      "The Sklml compiler will accept all OCaml options plus these ones:\n";
    match get_mode () with
    | None -> failwith "mode must be specified"
    | Some m ->
      ({
        mode = m;
        pp_cmd = mk_pp_cmd (get_pp ());
        ocamlc_options = get_annons ();
       }, get_debug ())
  with
  | Arg.Bad _ -> fail "invalid usage, see help using -help option"
  | Arg.Help m -> print_string m; exit 0
  | Failure m -> fail ("invalid usage, " ^ m)
;;

let dprintf b =
  (if b then Printf.fprintf else Printf.ifprintf) stdout
;;

let main () =
  let (compile_opts, dbg) = parse_argv Sys.argv in
  let compile_cmd = mk_command compile_opts in
  dprintf dbg "will run : %s\n" compile_cmd;
  let ret = Sys.command compile_cmd in
  dprintf dbg
    (if ret = 0
     then format_of_string "compilation finished\n"
     else format_of_string "there were errors\n");
  exit ret
;;

let _ = main ();;
